import base64
GXEou=Exception
GXEoH=True
GXEof=str
GXEol=len
GXEoA=bytes
GXEoQ=False
GXEoW=int
GXEoh=range
GXEoR=ValueError
GXEos=object
import json
import logging
import os
import sys
import traceback
from typing import Dict
from localstack import config as localstack_config
from localstack.constants import ENV_PRO_ACTIVATED
from localstack.utils.common import FileMappedDocument,load_file,md5,now_utc
from localstack.utils.common import safe_requests as requests
from localstack.utils.common import str_insert,str_remove,to_bytes,to_str
from localstack_ext import __version__,config
from localstack_ext.bootstrap.auth import get_auth_cache
from localstack_ext.bootstrap.decryption import DecryptionHandler,init_source_decryption
from localstack_ext.config import ROOT_FOLDER
ENV_PREPARED={}
MAX_KEY_CACHE_DURATION_SECS=60*60*24
LOG=logging.getLogger(__name__)
ENV_LOCALSTACK_API_KEY="LOCALSTACK_API_KEY"
class KeyActivationError(GXEou):
 pass
class CachedKeyError(KeyActivationError):
 pass
class InvalidKeyError(KeyActivationError):
 pass
class InvalidDecryptionKeyError(KeyActivationError):
 pass
def read_api_key(raise_if_missing=GXEoH):
 key=(os.environ.get(ENV_LOCALSTACK_API_KEY)or "").strip()
 if not key and raise_if_missing:
  raise GXEou("Unable to retrieve API key. Please configure $%s in your environment"%ENV_LOCALSTACK_API_KEY)
 return key
def truncate_api_key(api_key:GXEof):
 return '"%s..."(%s)'%(api_key[:3],GXEol(api_key))
def fetch_key()->GXEoA:
 api_key=read_api_key()
 if api_key=="test":
  return b"test"
 elif api_key=="enterprise":
  LOG.debug("Looking for a cached enterprise key, skipping online activation.")
  key_base64=load_cached_key(api_key=api_key,check_timeout=GXEoQ)
  decoded_key=base64.b64decode(key_base64)
  return decoded_key
 data={"api_key":api_key,"version":__version__}
 try:
  logging.getLogger("py.warnings").setLevel(logging.ERROR)
  result=requests.post("%s/activate"%config.API_URL,json.dumps(data),verify=GXEoQ)
  if result.status_code>=400:
   content=result.content
   content_type=result.headers.get("Content-Type")
   if result.status_code==403:
    message=json.loads(to_str(content))["message"]
    raise InvalidKeyError("Activation key %s is invalid or expired! Reason: %s"%(truncate_api_key(api_key),message))
   raise KeyActivationError('Received error activating key (code %s): ctype "%s" - %s'%(result.status_code,content_type,content))
  key_base64=json.loads(to_str(result.content))["key"]
  cache_key_locally(api_key,key_base64)
 except InvalidKeyError:
  raise
 except GXEou as e:
  if log_license_issues():
   api_key=GXEof(api_key_configured()or "")
   LOG.warning("Error activating API key %s: %s %s"%(truncate_api_key(api_key),e,traceback.format_exc()))
   LOG.warning("Looking for cached key as fallback...")
  key_base64=load_cached_key(api_key)
 finally:
  logging.getLogger("py.warnings").setLevel(logging.WARNING)
 decoded_key=base64.b64decode(key_base64)
 return decoded_key
def get_key_cache()->FileMappedDocument:
 return FileMappedDocument(os.path.join(localstack_config.dirs.cache,"key.json"),mode=0o600)
def cache_key_locally(api_key,key_b64):
 timestamp=GXEof(GXEoW(now_utc()))
 key_raw=to_str(base64.b64decode(key_b64))
 for i in GXEoh(GXEol(timestamp)):
  key_raw=str_insert(key_raw,i*2,timestamp[i])
 key_b64=to_str(base64.b64encode(to_bytes(key_raw)))
 cache=get_key_cache()
 cache.update({"timestamp":GXEoW(timestamp),"key_hash":md5(api_key),"key":key_b64})
 cache.save()
def load_cached_key(api_key:GXEof,check_timeout:bool=GXEoH)->GXEof:
 cache=get_key_cache()
 if not cache.get("key"):
  raise CachedKeyError("Could not find cached key")
 if cache.get("key_hash")!=md5(api_key):
  raise CachedKeyError("Cached key was created for a different API key")
 now=now_utc()
 if check_timeout and(now-cache["timestamp"])>MAX_KEY_CACHE_DURATION_SECS:
  raise CachedKeyError("Cached key expired")
 timestamp=GXEof(cache["timestamp"])
 key_raw=to_str(base64.b64decode(cache["key"]))
 for i in GXEoh(GXEol(timestamp)):
  assert key_raw[i]==timestamp[i]
  key_raw=str_remove(key_raw,i)
 key_b64=to_str(base64.b64encode(to_bytes(key_raw)))
 return key_b64
def enable_file_decryption(key:GXEoA):
 decryption_handler=DecryptionHandler(key)
 try:
  file_name=f"{ROOT_FOLDER}/localstack_ext/utils/common.py.enc"
  encrypted_file_content=load_file(file_name,mode="rb")
  file_content=decryption_handler.decrypt(encrypted_file_content)
  if "import" not in to_str(file_content):
   raise GXEoR("Decryption resulted in invalid python file!")
 except GXEou:
  raise InvalidDecryptionKeyError("Error while trying to validate decryption key!")
 init_source_decryption(decryption_handler)
def check_require_pro():
 if config.REQUIRE_PRO:
  LOG.error("Unable to activate API key, but $REQUIRE_PRO is configured - quitting.")
  sys.exit(1)
def prepare_environment():
 class OnClose(GXEos):
  def __exit__(self,*args,**kwargs):
   ENV_PREPARED["finalized"]=GXEoH
  def __enter__(self,*args,**kwargs):
   pass
 if not ENV_PREPARED.get("finalized"):
  try:
   key=fetch_key()
   if not key:
    raise GXEou("Unable to fetch and validate API key from environment")
   if to_str(key)!="test":
    enable_file_decryption(key)
    LOG.info("Successfully activated API key")
   else:
    LOG.info("Using test API key")
   os.environ[ENV_PRO_ACTIVATED]="1"
  except KeyActivationError as e:
   if log_license_issues():
    if LOG.isEnabledFor(level=logging.DEBUG):
     LOG.exception("exception while activating key")
    else:
     LOG.warning("error while activating key: %s",e)
   check_require_pro()
  except GXEou as e:
   if log_license_issues():
    LOG.warning("Unable to activate API key: %s %s"%(e,traceback.format_exc()))
   check_require_pro()
 return OnClose()
def log_license_issues():
 return api_key_configured()and localstack_config.is_env_not_false("LOG_LICENSE_ISSUES")
def api_key_configured():
 return read_api_key(raise_if_missing=GXEoQ)
def is_logged_in():
 return GXEoH if get_auth_cache().get("token")else GXEoQ
def get_auth_headers()->Dict:
 auth_cache=get_auth_cache()
 token=auth_cache.get("token")
 if token:
  if not token.startswith("%s "%auth_cache["provider"]):
   token="%s %s"%(auth_cache["provider"],token)
  return{"authorization":token}
 api_key=read_api_key(raise_if_missing=GXEoQ)
 if api_key:
  return{"ls-api-key":api_key,"ls-version":__version__}
 raise GXEou("Please log in first")
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
YnMzJ=bool
YnMzr=False
LOG=logging.getLogger(__name__)
def configure_systemd(revert:YnMzJ):
 from subprocess import CalledProcessError
 from localstack import config
 from localstack.utils.docker_utils import DOCKER_CLIENT,ContainerException
 from localstack.utils.run import is_linux,run,to_str
 if not is_linux():
  LOG.warning("Command only supported on GNU/Linux")
 try:
  container_ip=DOCKER_CLIENT.get_container_ip(config.MAIN_CONTAINER_NAME)
  container_network=DOCKER_CLIENT.get_network(config.MAIN_CONTAINER_NAME)
  if container_network=="default":
   container_network="bridge"
  network_inspect=DOCKER_CLIENT.inspect_network(container_network)
  network_interface=network_inspect["Options"].get("com.docker.network.bridge.name")
  network_interface=network_interface or f"br-{network_inspect['Id'][:12]}"
  if revert:
   cmd=["sudo","resolvectl","revert",network_interface]
   run(cmd,shell=YnMzr,print_error=YnMzr)
   LOG.info("Reverting DNS config completed!")
  else:
   cmd=["sudo","resolvectl","dns",network_interface,container_ip]
   run(cmd,shell=YnMzr,print_error=YnMzr)
   cmd=["sudo","resolvectl","domain",network_interface,"~amazonaws.com","~aws.amazon.com","~cloudfront.net","~localhost.localstack.cloud"]
   run(cmd,shell=YnMzr,print_error=YnMzr)
   LOG.info("Setting DNS config completed!")
 except ContainerException:
  if config.DEBUG:
   LOG.exception("Error while getting container information")
  LOG.warning("LocalStack container might not be running. Is container %s running?",config.MAIN_CONTAINER_NAME)
  LOG.warning("Is $MAIN_CONTAINER_NAME set correctly?")
 except CalledProcessError as e:
  LOG.warning("Error while configuring systemd-resolved: %s",to_str(e.output).strip())
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.deployment_utils import generate_default_name_without_stack
CcnOM=staticmethod
CcnOv=None
CcnOQ=all
CcnOK=super
CcnOf=str
CcnOd=classmethod
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
class SESTemplate(GenericBaseModel):
 @CcnOM
 def cloudformation_type():
  return "AWS::SES::Template"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("ses")
  template_name=self.props.get("Template",{}).get("TemplateName")
  tmpl_name=self.resolve_refs_recursively(stack_name,template_name,resources)
  templates=client.list_templates().get("TemplatesMetadata",[])
  template=[t for t in templates if t["Name"]==tmpl_name]
  return(template or[CcnOv])[0]
 def get_physical_resource_id(self,attribute=CcnOv,**kwargs):
  return self.props.get("Template",{}).get("TemplateName")
 def update_resource(self,new_resource,stack_name,resources):
  client=aws_stack.connect_to_service("ses")
  new_props=new_resource["Properties"]
  new_template=new_props.get("Template",{})
  template=client.get_template(TemplateName=new_template["TemplateName"])["Template"]
  if CcnOQ(template.get(attr,"")==new_template.get(attr,"")for attr in["SubjectPart","TextPart","HtmlPart"]):
   return
  return client.update_template(**new_props)
 @CcnOM
 def get_deploy_templates():
  return{"create":{"function":"create_template"},"delete":{"function":"delete_template","parameters":{"TemplateName":"TemplateName"}}}
class SESReceiptRuleSet(GenericBaseModel):
 @CcnOM
 def cloudformation_type():
  return "AWS::SES::ReceiptRuleSet"
 def get_physical_resource_id(self,attribute=CcnOv,**kwargs):
  return self.props.get("RuleSetName")
 def get_cfn_attribute(self,attribute):
  return CcnOK(SESReceiptRuleSet,self).get_cfn_attribute(attribute)
 @CcnOM
 def add_defaults(resource,stack_name:CcnOf):
  role=resource["Properties"]
  if not role.get("RuleSetName"):
   role["RuleSetName"]=generate_default_name_without_stack(resource["LogicalResourceId"])
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("ses")
  rule_set_name=self.props.get("RuleSetName")
  rule_set_name=self.resolve_refs_recursively(stack_name,rule_set_name,resources)
  rule_set=client.describe_receipt_rule_set(RuleSetName=rule_set_name)
  return rule_set or CcnOv
 @CcnOd
 def fetch_details(cls,rule_set_name):
  client=aws_stack.connect_to_service("ses")
  rule=client.describe_receipt_rule_set(RuleSetName=rule_set_name)
  return rule or CcnOv
 @CcnOM
 def get_deploy_templates():
  return{"create":{"function":"create_receipt_rule_set"},"delete":{"function":"delete_receipt_rule_set"}}
class SESReceiptRule(GenericBaseModel):
 @CcnOM
 def cloudformation_type():
  return "AWS::SES::ReceiptRule"
 def get_physical_resource_id(self,attribute=CcnOv,**kwargs):
  return self.props.get("Rule",{}).get("Name")
 def get_cfn_attribute(self,attribute):
  return CcnOK(SESReceiptRule,self).get_cfn_attribute(attribute)
 @CcnOd
 def fetch_details(cls,rule_set_name,rule_name):
  client=aws_stack.connect_to_service("ses")
  rule=client.describe_receipt_rule(RuleSetName=rule_set_name,RuleName=rule_name).get("Rule")
  return rule or CcnOv
 @CcnOM
 def add_defaults(resource,stack_name:CcnOf):
  rule=resource["Properties"]["Rule"]
  if not rule.get("Name"):
   rule["Name"]=generate_default_name_without_stack(resource["LogicalResourceId"])
 @CcnOM
 def get_deploy_templates():
  return{"create":{"function":"create_receipt_rule"},"delete":{"function":"delete_receipt_rule","parameters":{"RuleSetName":"RuleSetName","RuleName":lambda params,**kwargs:params["Rule"]["Name"]}}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
ElgCL=set
ElgCY=str
ElgCS=None
ElgCs=isinstance
ElgCM=list
ElgCO=Exception
import re
from email.header import Header
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from localstack.utils.common import now_utc
from localstack.utils.testutil import is_local_test_mode
from localstack_ext import config
from localstack_ext.bootstrap import smtplib_patched
LOG=logging.getLogger(__name__)
SENT_EMAILS=[]
EMAIL_BLACKLIST=ElgCL()
def is_smtp_configured():
 return config.SMTP_HOST
def get_canonical_email(email):
 email=re.sub(r"\s+","",ElgCY(email or ""))
 email=email.strip().lower()
 return email
def send_email(subject,message,recipients,from_email=ElgCS,from_name=ElgCS,smtp_host=ElgCS,smtp_user=ElgCS,smtp_pass=ElgCS,images={}):
 smtp_host=smtp_host or config.SMTP_HOST
 smtp_user=smtp_user or config.SMTP_USER
 smtp_pass=smtp_pass or config.SMTP_PASS
 from_email=from_email or config.SMTP_EMAIL
 from_name=from_name or "LocalStack"
 if not smtp_host:
  if is_local_test_mode():
   entry={"time":now_utc(),"smtp_host":smtp_host,"smtp_user":smtp_user,"smtp_pass":smtp_pass,"from_email":from_email,"from_name":from_name,"subject":subject,"message":message,"recipients":recipients}
   SENT_EMAILS.append(entry)
   return
  LOG.debug('SMTP settings not configured, skip sending email to "%s"'%recipients)
  return
 recipients=recipients if ElgCs(recipients,ElgCM)else[recipients]
 message=construct_message(subject,message,from_name,from_email,images=images)
 sign_message(message)
 for recipient in recipients:
  if recipient in EMAIL_BLACKLIST:
   LOG.debug("Skip sending email to receiver in blacklist: %s"%recipient)
   continue
  LOG.debug("Sending email to %s"%recipient)
  message["To"]=recipient
  s=smtplib_patched.SMTP(smtp_host)
  try:
   s.starttls()
  except ElgCO as e:
   LOG.debug("Unable to run STARTTLS command on SMTP connection: %s"%e)
  if smtp_user and smtp_pass:
   try:
    s.login(smtp_user,smtp_pass)
   except ElgCO as e:
    LOG.debug("Unable to run login/auth command against SMTP server, skipping: %s"%e)
  s.sendmail(from_email,[recipient],message.as_string())
  s.quit()
def sign_message(msg):
 pass
def is_email_address(value):
 return re.match(r"[^@]+@[^@]+\.[^@]+",value or "")
def construct_message(subject,message,from_name,from_email,images=ElgCS):
 result=MIMEText(message)
 if images:
  result=MIMEMultipart("related")
  msg_alternative=MIMEMultipart("alternative")
  result.attach(msg_alternative)
  msg_plain=MIMEText(message)
  msg_alternative.attach(msg_plain)
  msg_html=MIMEText(message,"html")
  msg_alternative.attach(msg_html)
  for image_id,image_bytes in images.items():
   msg_image=MIMEImage(image_bytes)
   msg_image.add_header("Content-ID","<%s>"%image_id)
   result.attach(msg_image)
 result["Subject"]=subject
 result["From"]=formataddr((ElgCY(Header(from_name,"utf-8")),from_email))
 return result
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
XHPFN=staticmethod
XHPFk=None
XHPFQ=super
XHPFv=Exception
XHPFE=len
XHPFW=str
XHPFc=classmethod
from localstack.utils.aws import aws_stack
class ElastiCacheSubnetGroup(GenericBaseModel):
 @XHPFN
 def cloudformation_type():
  return "AWS::ElastiCache::SubnetGroup"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("elasticache")
  groups=client.describe_cache_subnet_groups().get("CacheSubnetGroups",[])
  grp_name=self.resolve_refs_recursively(stack_name,self.props["CacheSubnetGroupName"],resources)
  result=[e for e in groups if e["CacheSubnetGroupName"]==grp_name]
  return(result or[XHPFk])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("CacheSubnetGroupName")
 @XHPFN
 def get_deploy_templates():
  return{"create":{"function":"create_cache_subnet_group","parameters":{"CacheSubnetGroupName":"CacheSubnetGroupName","CacheSubnetGroupDescription":"Description","SubnetIds":"SubnetIds","Tags":"Tags"}},"delete":{"function":"delete_cache_subnet_group","parameters":["CacheSubnetGroupName"]}}
class ElastiCacheReplicationGroup(GenericBaseModel):
 @XHPFN
 def cloudformation_type():
  return "AWS::ElastiCache::ReplicationGroup"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("elasticache")
  groups=client.describe_replication_groups().get("ReplicationGroups",[])
  grp_desc=self.resolve_refs_recursively(stack_name,self.props["ReplicationGroupDescription"],resources)
  result=[e for e in groups if e.get("Description")==grp_desc]
  return(result or[XHPFk])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("ReplicationGroupId")
 @XHPFN
 def get_deploy_templates():
  return{"create":{"function":"create_replication_group"},"delete":{"function":"delete_replication_group","parameters":["ReplicationGroupId"]}}
class ElastiCacheSecurityGroup(GenericBaseModel):
 @XHPFN
 def cloudformation_type():
  return "AWS::ElastiCache::SecurityGroup"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("elasticache")
  groups=client.describe_cache_security_groups().get("CacheSecurityGroups",[])
  grp_desc=self.resolve_refs_recursively(stack_name,self.props["Description"],resources)
  result=[e for e in groups if e["Description"]==grp_desc]
  return(result or[XHPFk])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("CacheSecurityGroupName")
 @XHPFN
 def get_deploy_templates():
  return{"create":{"function":"create_cache_security_group"},"delete":{"function":"delete_cache_security_group","parameters":["CacheSecurityGroupName"]}}
class ElastiCacheParameterGroup(GenericBaseModel):
 @XHPFN
 def cloudformation_type():
  return "AWS::ElastiCache::ParameterGroup"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("elasticache")
  groups=client.describe_cache_parameter_groups().get("CacheParameterGroups",[])
  grp_name=self.resolve_refs_recursively(stack_name,self.props["CacheParameterGroupName"],resources)
  result=[e for e in groups if e["CacheParameterGroupName"]==grp_name]
  return(result or[XHPFk])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("CacheParameterGroupName")
 @XHPFN
 def get_deploy_templates():
  def _param_values(params,**kwargs):
   props_dict=params.get("Properties")or{}
   return[{"ParameterName":n,"ParameterValue":v}for n,v in props_dict.items()]
  return{"create":[{"function":"create_cache_parameter_group","parameters":["CacheParameterGroupName","CacheParameterGroupFamily","Description","Tags"]},{"function":"modify_cache_parameter_group","parameters":{"CacheParameterGroupName":"CacheParameterGroupName","ParameterNameValues":_param_values}}],"delete":{"function":"delete_cache_parameter_group","parameters":["CacheParameterGroupName"]}}
class ElastiCacheCluster(GenericBaseModel):
 @XHPFN
 def cloudformation_type():
  return "AWS::ElastiCache::CacheCluster"
 def get_cfn_attribute(self,attribute_name):
  try:
   result=XHPFQ(ElastiCacheCluster,self).get_cfn_attribute(attribute_name)
   assert result is not XHPFk
   return result
  except XHPFv:
   props=self.props
   if attribute_name in["Port","Address"]:
    attribute_name="RedisEndpoint.%s"%attribute_name
   parts=attribute_name.split(".")
   candidates=["ConfigurationEndpoint","RedisEndpoint","Endpoint"]
   if parts[0]in candidates and XHPFE(parts)>1:
    key_name="ConfigurationEndpoint"
    if not props.get(key_name):
     self.state=self.fetch_details(props.get("ClusterName"))
    parent_obj=self.state.get(parts[0])or self.state.get(key_name)or{}
    result=parent_obj.get(parts[1])or props.get(parts[1])
    return XHPFW(result)
   return props.get(attribute_name)
 @XHPFc
 def fetch_details(cls,cluster_name):
  client=aws_stack.connect_to_service("elasticache")
  clusters=client.describe_cache_clusters(MaxRecords=500).get("CacheClusters",[])
  match=[c for c in clusters if c.get("CacheClusterId")==cluster_name]
  return(match or[XHPFk])[0]
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import json
BOAKU=staticmethod
BOAKY=str
BOAKb=None
BOAKX=int
BOAKi=Exception
BOAKT=list
from localstack import config
from localstack.services.awslambda import lambda_executors
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import retry,short_uid,to_str
from localstack.utils.testutil import map_all_s3_objects
from localstack_ext.services.cloudformation.service_models import(CUSTOM_RESOURCE_STATUSES,CUSTOM_RESOURCES_RESULT_POLL_TIMEOUT,CUSTOM_RESOURCES_RESULTS_BUCKET,LOG)
from localstack_ext.utils.aws import aws_utils
class CDKMetadata(GenericBaseModel):
 @BOAKU
 def cloudformation_type():
  return "AWS::CDK::Metadata"
 def fetch_state(self,stack_name,resources):
  return{"type":"AWS::CDK::Metadata"}
class CustomResource(GenericBaseModel):
 @BOAKU
 def cloudformation_type():
  return "AWS::CloudFormation::CustomResource"
 def fetch_state(self,stack_name,resources):
  resource_id=self.logical_resource_id
  result=(CUSTOM_RESOURCE_STATUSES.get(aws_stack.get_region(),{}).get(stack_name,{}).get(resource_id))
  return result
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.logical_resource_id
 @BOAKU
 def get_deploy_templates():
  def create_custom_resource(resource_id,resources,resource_type,func,stack_name,*args):
   resource=resources[resource_id]
   resource_props=resource["Properties"]
   service_token=resource_props.get("ServiceToken")
   if not service_token:
    LOG.warning("Missing ServiceToken attribute in custom resource: %s"%resource)
    return
   s3_client=aws_stack.connect_to_service("s3")
   s3_client.create_bucket(Bucket=CUSTOM_RESOURCES_RESULTS_BUCKET)
   result_key=short_uid()
   main_endpoint=lambda_executors.get_main_endpoint_from_container()
   endpoint_url="http://%s:%s"%(main_endpoint,config.get_edge_port_http())
   response_url=aws_stack.generate_presigned_url("put_object",Params={"Bucket":CUSTOM_RESOURCES_RESULTS_BUCKET,"Key":result_key},endpoint_url=endpoint_url)
   stack_arn=aws_stack.cloudformation_stack_arn(stack_name)
   request={"RequestType":"Create","ResponseURL":response_url,"StackId":stack_arn,"RequestId":short_uid(),"ResourceType":resource.get("Type"),"LogicalResourceId":resource_id,"ResourceProperties":resource_props}
   if BOAKY(service_token).startswith("arn:aws:lambda"):
    region_name=service_token.split(":")[3]
    function_name=aws_stack.lambda_function_name(service_token)
    lambda_client=aws_stack.connect_to_service("lambda",region_name=region_name)
    lambda_client.invoke(FunctionName=function_name,Payload=json.dumps(request))
   else:
    LOG.warning("Unsupported ServiceToken attribute in custom resource: %s"%service_token)
    return
   def fetch_result():
    return aws_utils.download_s3_object(CUSTOM_RESOURCES_RESULTS_BUCKET,result_key)
   result=BOAKb
   try:
    result=retry(fetch_result,retries=BOAKX(CUSTOM_RESOURCES_RESULT_POLL_TIMEOUT/2),sleep=2)
    result=json.loads(to_str(result))
   except BOAKi:
    bucket_objects=map_all_s3_objects(buckets=[CUSTOM_RESOURCES_RESULTS_BUCKET])
    LOG.info("Unable to fetch CF custom resource result from s3://%s/%s . Existing keys: %s"%(CUSTOM_RESOURCES_RESULTS_BUCKET,result_key,BOAKT(bucket_objects.keys())))
    raise
   region=aws_stack.get_region()
   status=CUSTOM_RESOURCE_STATUSES[region]=CUSTOM_RESOURCE_STATUSES.get(region,{})
   status_stack=status[stack_name]=status.get(stack_name,{})
   status_stack[resource_id]={"result":result}
   return result
  return{"create":{"function":create_custom_resource}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)
